package readerwriter;

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

import java.util.*;

/**
 * La classe FifoThreadSelector implementa un semplicissimo selezionatore di Thread (interfaccia
 * ThreadSelector) : un thread che viene messo in coda, viene risvegliato quando tutti quelli che
 * erano arrivati prima di lui a mettersi in coda sono stati risvegliati. La coda e' FIFO (First
 * In First Out), il primo ad essere arrivato e' anche il primo ad andarsene
 */
public class FifoThreadSelector implements ThreadSelector
{ private static final int DEFAULT_QUEUE_SIZE = 10;
  private int arrayHeader;
  private int arrayTail;
  private int arrayCapacity;
  private Semaphore sem;
  private SemaphoredThread[] arrayThread;

  public FifoThreadSelector()
  { this(DEFAULT_QUEUE_SIZE);
  }

  public FifoThreadSelector(int fifoCapacity)
  { arrayHeader = 0;
    arrayTail = 0;
    arrayCapacity = fifoCapacity;
    sem = new Semaphore();
    arrayThread = new SemaphoredThread[arrayCapacity];
  }




  /**
   * Inserisce un nuovo Thread in coda lista dei Thread in attesa; si preoccupa di metterlo
   * in wait su di se stesso in attesa di essere risvegliato. Il metodo e' thread safe
   *
   * @param newThread Thread che viene posto in stato di wait e inserito in coda
   * @param priority Non viene considerato
   */
  public void addThread(SemaphoredThread newThread, Object priority)
  { newThread.down(sem);
    addTail(newThread);
    synchronized(newThread)
    { newThread.up(sem);
      newThread.upWaitAll();
      try
      { newThread.wait();
      }
      catch(InterruptedException e)
      { e.printStackTrace();
      }
      newThread.downWaitAll();
    }
  }

  /**
   * Rimuove il Thread in testa alla lista dei Thread in attesa; si preoccupa di
   * risvegliare il Thread dal suo stato di wait. Il metodo e' thread safe
   *
   * @return Thread Il thread che e' stato risvegliato
   */
  public SemaphoredThread removeThread()
  { SemaphoredThread notifyThread = ((SemaphoredThread)Thread.currentThread());
    notifyThread.down(sem);
    Thread wakeThread = removeHeader();
    if (wakeThread != null)
    { synchronized(wakeThread)
      { wakeThread.notify();
      }
    }
    notifyThread.up(sem);
    return (SemaphoredThread)wakeThread;
  }


  /**
   * Inserisce nella coda oggetto della classe <code>Thread</code>
   *
   * @param tail oggetto della classe <code>Thread</code> da inserire nella coda
   */
  private void addTail(SemaphoredThread tail)
  { if (arrayThread[arrayTail] == null)
    { // La coda puo' gia' contenere un elemento valido
      arrayThread[arrayTail] = tail;
      arrayTail = (++arrayTail)%arrayCapacity;
    }
    else
    { // La testa e la coda coincidono, l'array e' saturo, va ridimensionato
      resizeArray();
      arrayThread[arrayTail] = tail;
      arrayTail = (++arrayTail)%arrayCapacity;
    }
  }


  /**
   * Seleziona dalla testa della coda un oggetto di classe <code>Thread</code>
   *
   * @return Thread oggetto della classe <code>Thread</code> da inserire nella coda
   */
  private SemaphoredThread removeHeader()
  { SemaphoredThread returnElement;

    if (arrayThread[arrayHeader] != null)
    { // La testa e' gia' occupata da un elemento valido, non deve essere incrementata
      returnElement = arrayThread[arrayHeader];
      arrayThread[arrayHeader] = null;
      if (arrayHeader != arrayTail)
        arrayHeader = (++arrayHeader)%arrayCapacity;
      return returnElement;
    }
    else
    { // La coda e' vuota
      return null;
    }
  }


  /**
   * Incremento delle dimensioni dell'array. Attenzione : non si tiene mai conto di eventuali
   * decrementi delle dimensioni dell'array, ma solo di incrementi; alla fine dell'operazione
   * l'indice arrayTail deve essere posizionato sulla prima posizione libera dell'array,
   * e l'indice arrayHeader sulla prima posizione occupata dell'array
   */
  private void resizeArray()
  { int newCapacity = arrayCapacity*2;
    int numElem = arrayCapacity;
    SemaphoredThread insElement;
    SemaphoredThread[] newArray = new SemaphoredThread[newCapacity];

    arrayCapacity = newCapacity;
    for(int index=0; index<numElem; index++)
    { insElement = removeHeader();
      if (insElement != null)
        newArray[index] = insElement;
      else
        break;
    }
    arrayThread = null;
    arrayThread = newArray;
    arrayTail = numElem;
    arrayHeader = 0;
  }


  public static void main(String[] args)
  { int i;
    TestFifoSelector[] test = new TestFifoSelector[10];
    SemaphoredThread[] threadTest = new SemaphoredThread[10];
    FifoThreadSelector selector = new FifoThreadSelector();

    for (i=0; i<8; i++)
    { test[i] = new TestFifoSelector("Test " + i, TestFifoSelector.ADD, selector);
      threadTest[i] = new SemaphoredThread(test[i]);
    }
    for ( ; i<10; i++)
    { test[i] = new TestFifoSelector("Test " + i, TestFifoSelector.REMOVE, selector);
      threadTest[i] = new SemaphoredThread(test[i]);
    }

    for (i=0; i<10; i++)
      threadTest[i].start();
  }

}

















