package readerwriter;

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

/**
 * La classe PriorityThreadSelector implementa un selezionatore di Thread (interfaccia ThreadSelector),
 * a priorita', costruendo per ognuna delle 10 priorita' possibili una selezionatore di Thread FIFO
 * (FifoThreadSelector), che dovra' gestire tutti i thread che hanno la stessa priorita'.
 * Quando un thread viene aggiunto, a seconda della sua priorita' viene aggiunto nella coda appropriata,
 * il primo thread risvegliato, e' il primo thread selezionabile in modalita' FIFO, nella prima coda non
 * vuota in ordine di priorita' (MIN_PRIORITY -> MAX_PRIORITY)
 */

public class PriorityThreadSelector implements ThreadSelector
{ public static final int MIN_PRIORITY = 0;
  public static final int MAX_PRIORITY = 10;

  private FifoThreadSelector[] fifoSelector;
  private Semaphore sem1;
  private Semaphore sem2;

  public PriorityThreadSelector()
  { fifoSelector = new FifoThreadSelector[MAX_PRIORITY];
    for (int index=MIN_PRIORITY; index<MAX_PRIORITY; index++)
      fifoSelector[index] = new FifoThreadSelector();
    sem1 = new Semaphore();
    sem2 = new Semaphore();
  }

  /**
   * Mettodo dell'interfaccia usato come wrapper per il metodo effettivo della classe che prende
   * un oggetto della classe <code>Integer</code> come oggetto per la definizione della priorita',
   * il metodo tentera' una conversione del parametro attuale in un oggetto di classe <code>Integer</code>,
   * se la conversione non puo' avere luogo, la priorita' sara' settata a zero.
   * Il metodo e' thread safe
   *
   * @param newThread Thread corrente che deve essere aggiunto nella lista dei thread in attesa
   * @param priority oggetto generico che definisce la priorita' da assegnare al Thread aggiunto
   */
  public void addThread(SemaphoredThread newThread, Object priority)
  { // La priorita' nel caso di questa classe concreta, e' rappresentata da un intero,
    // ma l'interfaccia esige un Object generico, al fine di poter trattare questa classe
    // nel suo aspetto implementativo, ma anche nell'astrazione dell'interfaccia che
    // implementa, conserviamo il metodo dell'interfaccia, cercando di convertire i
    // parametri attuali di questo metodo

    newThread.down(sem1);
    if (priority instanceof Integer)
      addThread(newThread, (Integer)priority);
    else
      addThread(newThread, new Integer(0));
    newThread.up(sem1);
  }

  /**
   * Inserisce un nuovo Thread nella lista dei Thread in attesa, a seconda della sua priorita' viene
   * inserito in una coda dove ci sono in attesa tutti i thread della sua stessa priorita'.
   * Metodo semanticamente equivalente a quello dell'interfaccia ThreadSelector, ma sintatticamente
   * specializzato, nella specificazione del parametro formale che definisce la priorita' che in questo
   * caso e' un oggetto di classe <code>Integer</code>. Il metodo e' thread safe
   *
   * @param newThread Thread corrente che deve essere aggiunto nella lista dei thread in attesa
   * @param priority Integer che ne identifica la priorita', l'intero (oggetto della classe
   * <code>Integer</code>) puo' andare da un minimo di MIN_PRIORITY ad un massimo di MAX_PRIORITY
   */
  public void addThread(SemaphoredThread newThread, Integer priority)
  { newThread.down(sem2);
    int indexPriority = priority.intValue();
    if (indexPriority < MIN_PRIORITY)
      indexPriority = MIN_PRIORITY;
    else if (indexPriority > MAX_PRIORITY)
      indexPriority = MAX_PRIORITY;
    // Attenzione!! Questa e' una regione critica, c'e' pericolo di inconsistenza dei dati!
    // Infatti nei test da dei problemi, ma non ho trovato delle soluzioni a proposito, se avete
    // suggerimenti => mailto:sungray@tiscalinet.it
    fifoSelector[indexPriority].addThread(newThread, null);
    newThread.up(sem2);
  }

  /**
   * Rimuove il Thread Thread dalla lista dei Thread in attesa; si preoccupa di
   * risvegliare il Thread dal suo stato di wait. Il metodo e' thread safe
   *
   * @return SemaphoredThread Il thread che e' stato selezionato dalla lista dei thread in attesa,
   * e che e' stato risvegliato; null se non c'erano thread in attesa al momento della chiamata
   */
  public SemaphoredThread removeThread()
  { SemaphoredThread notifyThread = (SemaphoredThread)Thread.currentThread();
    SemaphoredThread resumedThread = null;

    notifyThread.down(sem2);
    for (int i=MAX_PRIORITY-1; (resumedThread==null) && (i>=MIN_PRIORITY); i--)
    { resumedThread = fifoSelector[i].removeThread();
    }
    notifyThread.down(sem2);

    return resumedThread;
  }


  public static void main(String[] args)
  { int i;
    TestPrioritySelector[] test = new TestPrioritySelector[30];
    SemaphoredThread[] threadTest = new SemaphoredThread[30];
    PriorityThreadSelector selector = new PriorityThreadSelector();

    for (i=0; i<20; i++)
    { test[i] = new TestPrioritySelector("Test " + i, TestPrioritySelector.ADD, selector);
      threadTest[i] = new SemaphoredThread(test[i]);
    }
    for ( ; i<30; i++)
    { test[i] = new TestPrioritySelector("Test " + i, TestPrioritySelector.REMOVE, selector);
      threadTest[i] = new SemaphoredThread(test[i]);
    }

    for (i=0; i<30; i++)
      threadTest[i].start();
  }

}