package readerwriter;

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

/**
 * La classe simula la regolazione di un flusso di thread (SemaphoredThread), che necessitano di
 * accedere ad una risorsa, controllata dalla classe, in lettura/scrittura, attraverso i metodi
 * leggi/scrivi. La risorsa in questione, qui e' stata solo simulata; La classe in generale, dovrebbe
 * fungere da Mediator : E' direttamente dipendente dalla risorsa (e' una sua interfaccia verso l'esterno),
 * definisce le politiche di accesso alla risorsa, e la schedulazione dei flussi d'esecuzione,
 * attraverso l'uso di opportune classi concrete specializzate, che rispettivamente estendono la classe
 * <code>Guardian<code>, e implementano l'interfaccia <code>ThreadSelector</code>
 * @see <a href="ReaderWriterClass.jpg">Diagramma delle classi di tutta la struttura</a>
 */

public class ReaderWriter
{ Guardian resourceGuardian;
  Semaphore semReaders;
  Semaphore semWriters;

  /**
   * Inizializza lo stato interno dell'oggetto, stabilendo quale guardiano utilizzare per la risorsa
   * 
   * @param guard Oggetto di classe <code>Guardian</code> che definira' la politica di selezione dei
   * thread (SemaphoredThread) Lettori/Scrittori
   * 
   */
  public ReaderWriter(Guardian guard)
  { resourceGuardian = guard;
    semReaders = new Semaphore();
    semWriters = new Semaphore();
  }


  /**
   * Wrapper del metodo di lettura della risorsa, il quale controlla con l'ausilio del guardiano
   * l'accesso in lettura della stessa
   *
   * @param priority Oggetto di classe <code>Integer</code> che determina la priorita' dell'oprazione,
   * <B>ATTENZIONE</B> priorita' valutabile solo per quanto riguarda le diverse opearazioni di lettura
   * richieste
   */
  public void leggi(Integer priority) throws SemaphoredThreadRequireException
  { SemaphoredThread readerThread;
    Thread testThread = Thread.currentThread();

    if (testThread instanceof SemaphoredThread)
    { readerThread = (SemaphoredThread)testThread;
      // Sembra un assurdo, ma si deve proteggere l'atomicita' dell'utilizzo del Guardian :-)
      readerThread.down(semReaders);
      resourceGuardian.iniziaLettura(priority);
      readerThread.up(semReaders);
      try
      { // simulazione del tempo di lettura, qui andrebbe la chiamata al metodo di lettura
        // della risorsa protetta
        Thread.sleep((int)(Math.random()*5000));
      }
      catch(InterruptedException e)
      { e.printStackTrace();
      }
      readerThread.down(semReaders);
      resourceGuardian.fineLettura();
      readerThread.up(semReaders);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }


  /**
   * Wrapper del metodo di scrittura della risorsa, il quale controlla con l'ausilio del guardiano
   * l'accesso in scrittura della stessa
   *
   * @param priority Oggetto di classe <code>Integer</code> che determina la priorita' dell'oprazione,
   * <B>ATTENZIONE</B> priorita' valutabile solo per quanto riguarda le diverse opearazioni di scrittura
   * richieste
   */
  public void scrivi(Integer priority) throws SemaphoredThreadRequireException
  { SemaphoredThread writerThread;
    Thread testThread = Thread.currentThread();

    if (testThread instanceof SemaphoredThread)
    { writerThread = (SemaphoredThread)testThread;
      // Sembra un assurdo, ma si deve proteggere l'atomicita' dell'utilizzo del Guardian :-)
      writerThread.down(semWriters);
      resourceGuardian.iniziaScrittura(priority);
      writerThread.up(semWriters);
      try
      { // simulazione del tempo di lettura, qui andrebbe la chiamata al metodo di lettura
        // della risorsa protetta
        Thread.sleep((int)(Math.random()*5000));
      }
      catch(InterruptedException e)
      { e.printStackTrace();
      }
      writerThread.down(semWriters);
      resourceGuardian.fineScrittura();
      writerThread.up(semWriters);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }

  public static void main(String[] args)
  { int i;
    ReaderWriter wrapper = new ReaderWriter(new ReadersPreferedGuardian(new FifoThreadSelector(),
      new FifoThreadSelector()));
    TestReaderWriter[] wrapperTest = new TestReaderWriter[10];
    SemaphoredThread[] threadTest = new SemaphoredThread[10];

    for (i=0; i<8; i++)
    { wrapperTest[i] = new TestReaderWriter("Test " + i, TestReaderWriter.WRITER, wrapper);
      threadTest[i] = new SemaphoredThread(wrapperTest[i]);
    }
    for ( ; i<10; i++)
    { wrapperTest[i] = new TestReaderWriter("Test " + i, TestReaderWriter.READER, wrapper);
      threadTest[i] = new SemaphoredThread(wrapperTest[i]);
    }

    for (i=0; i<10; i++)
      threadTest[i].start();
  }
}

