package readerwriter;

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

/**
 * La classe definisce le priorita' di accesso alla risorsa condivisa, da parte di Lettori e Scrittori,
 * preferendo i lettori : Ovviamente i lettori possono accedere tutti allo stesso momento alla risorsa,
 * ma se ci sono lettori sulla risorsa, l'accesso agli scrittori e' vietato, solo quando non esistono
 * piu' lettori, l'ultimo lettore, assegnera' la risorsa ad uno scrittore
 */

public class ReadersPreferedGuardian extends Guardian
{ private ThreadSelector readers;
  private ThreadSelector writers;
  // Variabile che indica se attualmente un thread sta accedendo alla risorsa il scrittura
  private boolean writerIsOn;
  private Semaphore sem;

  /**
   * Inizializza allo stato di default le variabili private della classe
   */
  public ReadersPreferedGuardian()
  { super();
    readers = new FifoThreadSelector();
    writers = new FifoThreadSelector();
    sem = new Semaphore();
    writerIsOn = false;
  }

  /**
   * Inizializza gli attributi privati della classe con i parametri attuali
   *
   * @param readers ThreadSelector che definira' la politica di risveglio dei thread in lettura
   * @param writers ThreadSelector che definira' la politica di risveglio dei thread in scrittura
   */
  public ReadersPreferedGuardian(ThreadSelector readers, ThreadSelector writers)
  { super();
    this.readers = readers;
    this.writers = writers;
    sem = new Semaphore();
    writerIsOn = false;
  }

  //----------------------------------------------------------------------------------------------------------

  /**
   * <p>Determina l'inizio dell'operazione di lettura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>I lettori hanno la priorita', quindi anche se ci sono lettori in attesa, i lettori possono
   * accedere, contemporaneamente vista la loro natura di lettori, alla risorsa. Ovviamente se c'e'
   * gia' uno scrittore che sta utilizzando la risorsa, allora il/i lettori dovranno attendere
   *
   * @param priority L'ogetto che rappresenta la priorita' da attribuire all'operazione
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void iniziaLettura(Object priority) throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread readerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { readerThread = (SemaphoredThread)tempThread;
      readerThread.down(sem);
      incrNumReaders();
      if (writerIsOn)
        readers.addThread((SemaphoredThread)readerThread, priority);
      readerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }


  /**
   * <p>Determina la fine dell'operazione di lettura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>La lettura della risorsa e' terminata, se il lettore attuale e' l'ultimo lettore della risorsa,
   * non ci sono altri lettori che stanno utilizzando la risorsa, e c'e' almeno uno scrittore che e'
   * in coda per accedere alla risorsa, questo viene selezionato
   *
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void fineLettura() throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread readerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { readerThread = (SemaphoredThread)tempThread;
      readerThread.down(sem);
      int numScrittori = getNumWriters();
      int numLettori = getNumReaders();
      if ((numLettori == 1) && (numScrittori > 0))
      { // Uno scrittore viene rimosso dalla coda e autorizzato ad accedere alla risorsa
        writers.removeThread();
      }
      decrNumReaders();
      readerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }


  /**
   * <p>Determina l'inizio dell'operazione di scrittura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>Se esiste qualcuno che sta leggendo o scrivendo sulla risorsa, l'operazione si interrompe,
   * in attesa di ricevere l'autorizzazione di procedere oltre
   *
   * @param priority L'ogetto che rappresenta la priorita' da attribuire all'operazione
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void iniziaScrittura(Object priority) throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread writerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { writerThread = (SemaphoredThread)tempThread;
      writerThread.down(sem);
      int numScrittori = getNumWriters();
      int numLettori = getNumReaders();
      incrNumWriters();
      if ((numScrittori>0) || (numLettori>0))
      { // Se esiste qualcuno che sta leggendo o scrivendo, dobbiamo metterci in attesa
        writers.addThread(writerThread, priority);
      }
      writerIsOn = true;
      writerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }


  /**
   * <p>Determina la fine dell'operazione di scrittura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>La scrittura della risorsa e' terminata, l'uso della risorsa ad uno scrittore,
   * potra' essere autorizzata, solo es esclusivamente se non esistono lettori che stanno attualmente
   * utilizzando la risorsa
   *
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void fineScrittura() throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread writerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { writerThread = (SemaphoredThread)tempThread;
      writerThread.down(sem);
      int numLettori = getNumReaders();
      if (numLettori == 0)
      { // Non esistono lettori, allora un altro scrittore puo' essere selezionato per accedere alla risorsa
        writers.removeThread();
      }
      else
      { // Il thread attuale stava utilizzando in scrittura la risorsa, questo vuol dire che i thread
        // che sono in attesa di leggere devono essere risvegliati
        while (readers.removeThread() != null)
        { }
      }
      decrNumWriters();
      writerIsOn = false;
      writerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }

  //----------------------------------------------------------------------------------------------------------


}