package readerwriter;

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

import java.util.*;

/**
 * Semaforo per la protezione in mutua esclusione di una risorsa condivisa da piu' thread,
 * i quali, per accedere alla risorsa devono eseguire l'opearazione di <code>down()</code>
 * sul semaforo, il quale in base al suo stato (e quindi agli altri thread presenti o meno),
 * quando il thread potra' proseguire in tranquillita', senza preoccuparsi di situazioni
 * critiche; una volta terminato l'utilizzo della risorsa, il thread dovra' informare il
 * semaforo eseguendo sullo stesso l'oprazione di <code>up()</code>
 *
 * <blockquote>
 * <pre>
 * private Semaphore sem = new Semaphore();
 * ...
 * sem.down(); // da questo punto in poi si puo' entrare nella regione critica
 * // REGIONE CRITICA
 * sem.up(); // il thread informa il semaforo, che sta' uscendo dalla regione critica
 * </pre>
 * </blockquote>
 */

public class Semaphore
{ private int numThread;
  private ArrayList listThread;

  /**
   * Inizializza allo stato di default lo stato interno
   */
  public Semaphore()
  { numThread = 0;
    listThread = new ArrayList();
  }

  /**
   * Esegue la classica operazione di <B>P</B> su di un semaforo, che ha termine quando nessun thread e'
   * nella regione critica, e quando, secondo i turni di arrivo delle richieste di <B>P</B> dei thread
   * (gestite in maniera FIFO), il thread attuale avra' il diritto di accedere alla regione critica
   *
   * @return Semaphore Il semaforo sul quale si sta operando
   */
  public synchronized Semaphore down()
  { Thread newThread = Thread.currentThread();
    if (listThread.indexOf(newThread) >= 0)
    { // Il thread non posside gia' la chiave del semaforo
      numThread++;
      listThread.add(Thread.currentThread());
      if (numThread > 1)
      { try
       { this.wait();
       }
       catch (InterruptedException e)
       { e.printStackTrace();
       }
      }
    }
    return this;
  }

  /**
   * Esegue la classica operazione di <B>V</B> su di un semaforo, informando il semaforo, che il thread
   * sta per lasciare la regione critica
   *
   * @return Semaphore Il semaforo sul quale si sta operando
   */
  public synchronized Semaphore up()
  { int indexThread = listThread.indexOf(Thread.currentThread());
    if (indexThread != -1)
    { if (numThread > 1)
      this.notify();
      listThread.remove(indexThread);
      numThread--;
      return this;
    }
    return null;
  }

}