package readerwriter;

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

 import java.util.*;

/**
 * La classe implemnta uno speciale tipo di thread, il cui flusso di esecuzione viene assoggettato a uno
 * o piu' semafori anche innestati fra di loro. La classe permette la gestione dei semafori, consentendo
 * di ricordare tutti i semafori dei quali il thread possiede la chiave, di richiedere la chiave di
 * un singolo thread, richiedere la chiave di tutti i semafori in SOSPENSIONE, rilasciare la chiave
 * di un semaforo, rilasciare temporaneamente la chiave di tutti i semafori in memoria, e quindi portarli
 * nello stato di SOSPENSIONE e rilasciare definitivamente la chiave di tutti i semafori in memoria
 */

public class SemaphoredThread extends Thread
{ private static final int INITIAL_CAPACITY = 10;
  private ArrayList semaphores;
  private Object[] tempWaitSemaphores = null;

  //----------------------------------------------------------------------------------------------------------

  /**
   * Alloca un nuovo oggetto <code>SemaphoredThread</code>
   */
  public SemaphoredThread()
  { super();
    initSemaphoresList(0);
  }

  /**
   * Alloca un nuovo oggetto <code>SemaphoredThread</code>
   */
  public SemaphoredThread(Runnable target)
  { super(target);
    initSemaphoresList(0);
  }

  /**
   * Alloca un nuovo oggetto <code>SemaphoredThread</code>
   */
  public SemaphoredThread(Runnable target, String name)
  { super(target, name);
    initSemaphoresList(0);
  }

  /**
   * Alloca un nuovo oggetto <code>SemaphoredThread</code>
   */
  public SemaphoredThread(String name)
  { super(name);
    initSemaphoresList(0);
  }

  /**
   * Alloca un nuovo oggetto <code>SemaphoredThread</code>
   */
  public SemaphoredThread(ThreadGroup group, Runnable target)
  { super(group, target);
    initSemaphoresList(0);
  }

  /**
   * Alloca un nuovo oggetto <code>SemaphoredThread</code>
   */
  public SemaphoredThread(ThreadGroup group, Runnable target, String name)
  { super(group, target, name);
    initSemaphoresList(0);
  }

  /**
   * Alloca un nuovo oggetto <code>SemaphoredThread</code>
   */
  public SemaphoredThread(ThreadGroup group, String name)
  { super(group, name);
    initSemaphoresList(0);
  }

  //----------------------------------------------------------------------------------------------------------

  /**
   * Richiede la chiave su uno specifico semaforo
   *
   * @param sem Semaforo del quale si richiede la chiave
   */
  public void down(Semaphore sem)
  { if (tempWaitSemaphores == null)
    { // Non si possono richiedere chiavi di altri semafori, quando si e' in attesa
      semaphores.add(sem);
      sem.down();
    }
  }

  /**
   * Si richiede il rilascio di uno specifico semaforo
   *
   * @param sem Semaforo del quale si richiede il rilascio della chiave
   */
  public void up(Semaphore sem)
  { if (tempWaitSemaphores == null)
    { // Non si possono rilaciare chiavi di semafori quando si e' in attesa
      sem.up();
      semaphores.remove(sem);
    }
  }

  /**
   * Si richiede il rilascio temporaneo di tutte le chiavi dei semafori in possesso del thread
   */
  public void upWaitAll()
  { if (tempWaitSemaphores == null)
    { tempWaitSemaphores = semaphores.toArray();
      int arrayLength = tempWaitSemaphores.length;

      for (int i=0; i<arrayLength; i++)
      { ((Semaphore)tempWaitSemaphores[i]).up();
      }
    }
  }


  /**
   * Si richiede il ripristino di tutte le chiavi in sospeso, che prima erano in possesso del thread
   */
  public void downWaitAll()
  { if (tempWaitSemaphores != null)
    { int arrayLength = tempWaitSemaphores.length;

      for (int i=0; i<arrayLength; i++)
      { ((Semaphore)tempWaitSemaphores[i]).down();
      }
      tempWaitSemaphores = null;
    }
  }


  /**
   * si richiede il rilascio definitivo di tutti i semafori in possesso del thread
   */
  public void upAll()
  { if (tempWaitSemaphores == null)
    { int listLength = semaphores.size();

      for (int i=0; i<listLength; i++)
      { ((Semaphore)semaphores.remove(0)).up();
      }
    }
  }

  //----------------------------------------------------------------------------------------------------------

  private void initSemaphoresList(int numElem)
  { if (numElem == 0)
      numElem = this.INITIAL_CAPACITY;
    semaphores = new ArrayList(numElem);
  }
}