package readerwriter;

/**
 * <p>Title:        ReaderWriter
 * <p>Description:  Progettazione di un mini framework per la gestione generale
 * di risorse condivise in maniera concorrente in lettura/scrittura
 *
 * <p>Copyright:    Copyright (c) 2001
 * <p>Company:      Universita' degli studi di Milano (Facolta' di Informatica)
 *
 * @author Gabriele "SunGray" Lana
 * @version 1.0
 */

/**
 * La classe definisce le priorita' di accesso alla risorsa condivisa, da parte di Lettori e Scrittori,
 * preferendo gli scrittori : Se uno scrittore tenta di accedere alla risorsa, nuovi lettori non sono
 * ammessi sulla risorsa, i lettori invece che stanno utilizzando la risorsa, vengono lasciati uscire.
 * In piu' se un lettore scrittore, finito l'utilizzo della risorsa, dovra' decidere a chi, fra un lettore
 * ed uno scrittore, assegnare la risorsa, preferira' quest'ultimo.
 */

public class WritersPreferedGuardian extends Guardian
{ private ThreadSelector readers;
  private ThreadSelector writers;
  private Semaphore sem;

  /**
   * Inizializza allo stato di default le variabili private della classe
   */
  public WritersPreferedGuardian()
  { super();
    readers = new FifoThreadSelector();
    writers = new FifoThreadSelector();
    sem = new Semaphore();
  }

  /**
   * Inizializza gli attributi privati della classe con i parametri attuali
   *
   * @param readers ThreadSelector che definira' la politica di risveglio dei thread in lettura
   * @param writers ThreadSelector che definira' la politica di risveglio dei thread in scrittura
   */
  public WritersPreferedGuardian(ThreadSelector readers, ThreadSelector writers)
  { super();
    this.readers = readers;
    this.writers = writers;
    sem = new Semaphore();
  }

  //----------------------------------------------------------------------------------------------------------

  /**
   * <p>Determina l'inizio dell'operazione di lettura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>Se esistono scrittori in attesa, il lettore si mette in pausa, in attesa che tutti
   * gli scrittori abbiano finito il loro compito. Se esistono altri lettori sulla risorsa,
   * l'opearazione procede senza interruzioni
   *
   * @param priority L'ogetto che rappresenta la priorita' da attribuire all'operazione
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void iniziaLettura(Object priority) throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread readerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { readerThread = (SemaphoredThread)tempThread;
      readerThread.down(sem);
      int numScrittori = getNumWriters();
      incrNumReaders();
      if (numScrittori>0)
      { // Preferiamo gli scrittori, se ci sono scrittori in attesa come in questo caso, ci mettiamo
        // in pausa in attesa che tutti gli scrittori abbiano finito il loro compito
        readers.addThread((SemaphoredThread)readerThread, priority);
      }
      readerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }


  /**
   * <p>Determina la fine dell'operazione di lettura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>La lettura della risorsa e' terminata, se esistono scrittori in attesa, si preferira' co
   * nsentire loro l'uso della risorsa, prima che ad eventuali altri lettori
   *
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void fineLettura() throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread readerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { readerThread = (SemaphoredThread)tempThread;
      readerThread.down(sem);
      int numScrittori = getNumWriters();
      if (numScrittori>0)
      { // Preferiamo gli scrittori, se ci sono scrittori in attesa svegliamo uno di loro
        writers.removeThread();
      }
      else if (getNumReaders() > 1)
        // In alternativa se ci sono dei lettori oltre noi, ne svegliamo uno di quelli
        readers.removeThread();
      decrNumReaders();
      readerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }


  /**
   * <p>Determina l'inizio dell'operazione di scrittura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>Se esiste qualcuno che sta leggendo o scrivendo sulla risorsa, l'operazione si interrompe,
   * in attesa di ricevere l'autorizzazione di procedere oltre
   *
   * @param priority L'ogetto che rappresenta la priorita' da attribuire all'operazione
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void iniziaScrittura(Object priority) throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread writerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { writerThread = (SemaphoredThread)tempThread;
      writerThread.down(sem);
      int numScrittori = getNumWriters();
      int numLettori = getNumReaders();
      incrNumWriters();
      if ((numScrittori>0) || (numLettori>0))
      { // Se esiste qualcuno che sta leggendo o scrivendo, dobbiamo metterci in attesa
        writers.addThread((SemaphoredThread)writerThread, priority);
      }
      writerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }


  /**
   * <p>Determina la fine dell'operazione di scrittura da parte di un thread
   * <p><b>ATTENZIONE</b> : L'operazione e' thread safe
   *
   * <p>La scrittura della risorsa e' terminata, si preferira' anche in questo caso consentire
   * l'uso della risorsa agli scrittori, piuttosto che ad eventuali altri lettori
   *
   * @exception SemaphoredThreadRequireException Eccezione lanciata se il thread che sta per eseguire il
   * metodo non e' un Thread di tipo <code>SemaphoredThread</code>
   */
  public void fineScrittura() throws SemaphoredThreadRequireException
  { // Qui non esiste il problema della sincronizzazione perche' stiamo lavorando
    // con variabili locali, quindi possiamo procedere in tutta tranquillita'
    SemaphoredThread writerThread;
    Thread tempThread = Thread.currentThread();
    if (tempThread instanceof SemaphoredThread)
    { writerThread = (SemaphoredThread)tempThread;
      writerThread.down(sem);
      int numScrittori = getNumWriters();
      int numLettori = getNumReaders();
      if (numScrittori>1)
      { // Esiste uno scrittore oltre all'attuale, che e' in attesa della risorsa, dobbiamo preferirlo
        writers.removeThread();
      }
      else if (numLettori>0)
      { // Ci sono solo lettori in attesa, a questo punto li svegliamo tutti
        while (readers.removeThread() != null)
        { }
      }
      decrNumWriters();
      writerThread.up(sem);
    }
    else
      throw new SemaphoredThreadRequireException("E' necessario un flusso d'esecuzione semaforizzato");
  }

  //----------------------------------------------------------------------------------------------------------


}